; hacker port firmware
; version 0.6.0 date 2005-04-15
; part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
; Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

	TITLE		"I2C procedures for hackerport firmware"

	include		"P18F442.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		DEBUG_LOG



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

I2C_BIT_TIME	res	1		; bit time of I2C (I2C_BIT_TIME-1 -> SSPADD)

I2C_BUSY_CNT	res	1		; number of times the I2C has been busy
#define		I2C_BUSY_MAX	100	; maximum number of times before reinitialization

I2C_BITS	res	1		; bits for I2C module
#define		I2C_BIT_STARTED		I2C_BITS,0	; if a start condition was output



;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; I2C bus initialization
;-------------------------------------------------------------------------------------------
I2C_INIT:
	movlw	100			; 100kHz
	movwf	I2C_BIT_TIME
I2C_REINIT:
; disable MSSP
	clrf	SSPCON1
; set up I2C in master mode
	movlw	b'10000000'		; no slew rate control, SMBus off, clear status flags
	movwf	SSPSTAT
	movlw	b'00101000'		; no collision yet, enable MSSP, master mode
	movwf	SSPCON1
	movlw	b'00000000'		; no acknowledge yet, recv off, stop off, repeated start off, start off
	movwf	SSPCON2
	decf	I2C_BIT_TIME,W		; set speed
	movwf	SSPADD
; I2C has never been busy
	clrf	I2C_BUSY_CNT
; I2C is not started
	bcf	I2C_BIT_STARTED
; done
	return

;-------------------------------------------------------------------------------------------
; check if I2C bus is idle
; input:	-
; output:	STATUS,C set if I2C busy
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_IDLE_CHECK:
	movlw	5			; try 5 times
	movwf	D1_REG
; check if busy
I2C_IDLE_CHECK_LOOP:
	btfsc	SSPSTAT,R_W		; transmit in progress?
	bra	I2C_IDLE_CHECK_BUSY
	movf	SSPCON2,W		; some other operation in progress?
	andlw	0x1F
	bnz	I2C_IDLE_CHECK_BUSY
; return "idle"
	clrf	I2C_BUSY_CNT		; clear busy counter
	bcf	STATUS,C
	return
; try a few times
I2C_IDLE_CHECK_BUSY:
	decfsz	D1_REG
	bra	I2C_IDLE_CHECK_LOOP
; I2C is busy, count this to recognize stuck bus
	incf	I2C_BUSY_CNT
	movlw	I2C_BUSY_MAX
	cpfslt	I2C_BUSY_CNT
	bra	I2C_IDLE_CHECK_REINIT
; return "not idle"
	bsf	STATUS,C
	return
; bus has been busy too often in a row - reinitialize
I2C_IDLE_CHECK_REINIT:
	call	I2C_REINIT
; return "idle"
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; set speed of I2C bus
; input:	A1_REG		bit time in 100ns
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, A1_REG, D1_REG
;-------------------------------------------------------------------------------------------
I2C_SPEED:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; bit time 0 is impossible -> change to 1
	movf	A1_REG
	btfsc	STATUS,Z
	incf	A1_REG
; save new bit time
	movf	A1_REG,W
	movwf	I2C_BIT_TIME
; renitialize I2C bus
	call	I2C_REINIT
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; output a start condition on I2C bus
; input:	-
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_STA:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; output start or restart condtition
	btfss	I2C_BIT_STARTED
	bsf	SSPCON2,SEN
	btfsc	I2C_BIT_STARTED
	bsf	SSPCON2,RSEN
; start condition has now been output
	bsf	I2C_BIT_STARTED
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; start writing a byte to I2C bus
; input:	A1_REG		the byte to write
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_WRITE_START:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; write byte
	movf	A1_REG,W
	movwf	SSPBUF
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; get acknowledge state of write to I2C bus
; input:	-
; output:	A1_REG		0 if no acknowledge was received
;		STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, A1_REG, D1_REG
;-------------------------------------------------------------------------------------------
I2C_WRITE_RESULT:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	clrf	A1_REG
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; get acknowledge state
	clrf	A1_REG
	btfss	SSPCON2,ACKSTAT
	incf	A1_REG
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; start reading a byte from I2C bus
; input:	-
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_READ_START:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; initiate read
	bsf	SSPCON2,RCEN
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; get byte read from I2C bus
; input:	-
; output:	A1_REG		the byte received
;		STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, A1_REG, D1_REG
;-------------------------------------------------------------------------------------------
I2C_READ_RESULT:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	setf	A1_REG
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; get received byte
	movf	SSPBUF,W
	movwf	A1_REG
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; output an acknowledge on I2C bus
; input:	-
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_ACK:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; output acknowledge
	bcf	SSPCON2,ACKDT
	bsf	SSPCON2,ACKEN
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; output a not-acknowledge on I2C bus
; input:	-
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_NAK:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; output not-acknowledge
	bsf	SSPCON2,ACKDT
	bsf	SSPCON2,ACKEN
; done
	bcf	STATUS,C
	return

;-------------------------------------------------------------------------------------------
; output a stop condition on I2C bus
; input:	-
; output:	STATUS,C	set if I2C is busy and command was not executed
; changes:	WREG, STATUS, D1_REG
;-------------------------------------------------------------------------------------------
I2C_STO:
; quit if I2C bus is busy
	call	I2C_IDLE_CHECK
	btfsc	STATUS,C
	return
; ignore command if no start condition was output on I2C
	bcf	STATUS,C
	btfss	I2C_BIT_STARTED
	return
; output stop condition
	bsf	SSPCON2,PEN
; start condition has not yet been output after this stop condition
	bcf	I2C_BIT_STARTED
; done
	bcf	STATUS,C
	return



;===========================================================================================
; global variables
;===========================================================================================



;===========================================================================================
; global procedures
;===========================================================================================

	global		I2C_INIT
	global		I2C_SPEED, I2C_STA, I2C_ACK, I2C_NAK, I2C_STO
	global		I2C_WRITE_START, I2C_WRITE_RESULT, I2C_READ_START, I2C_READ_RESULT



	end

