; hacker port firmware
; version 0.6.0 date 2005-04-15
; part of the hacker port project - http[s]://1stein.schuermans.info/hackerport/
; Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
; Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html

	TITLE		"message processing for hackerport firmware"

	include		"P18F442.INC"

	include		"IoPins.inc"
	include		"Bits.inc"
	include		"Version.inc"



;===========================================================================================
; imported variables
;===========================================================================================
	extern		BITS0
	extern		A1_REG, B1_REG, C1_REG, D1_REG
	extern		A2_REG, B2_REG, C2_REG, D2_REG
	extern		A4_REG
	extern		PROC_PTR, PROC_LEN
	extern		USB_Sx_CNT, USB_Sx_BUF



;===========================================================================================
; imported procedures
;===========================================================================================
	extern		EE_GET_ID, EE_SET_ID
	extern		SER_SET_BAUD, SER_SEND
	extern		I2C_SPEED, I2C_STA, I2C_ACK, I2C_NAK, I2C_STO
	extern		I2C_WRITE_START, I2C_WRITE_RESULT, I2C_READ_START, I2C_READ_RESULT
	extern		DEBUG_LOG



;===========================================================================================
accessram	UDATA_ACS		; access RAM
;===========================================================================================

; private bit variables
MSG_BITS	res		1
#define		MSG_SET_ID_LOCK		MSG_BITS,0	; setting identification is disabled
#define		MSG_BITS_I2C_PART2	MSG_BITS,1	; second part of some I2C command


;===========================================================================================
prog		CODE			; normal code
;===========================================================================================

;-------------------------------------------------------------------------------------------
; message processing initialization
;-------------------------------------------------------------------------------------------
MSG_INIT:
; initialize variables
	bcf	MSG_SET_ID_LOCK		; setting idnetification is allowed
	bcf	MSG_BITS_I2C_PART2	; not at second part of some I2C command
; done
	return

;-------------------------------------------------------------------------------------------
; process messages in packet
; input/output:	PROC_PTR	address of next byte to process
;		PROC_LEN	length of data to process
; changes:	PROC_PTR, PROC_LEN, FSR0, FSR1, PROD, X?_REG, WREG
; processes all messages in packet until a message cannot be processed
;-------------------------------------------------------------------------------------------
MSG_PROC:
; load address of next byte to process into FSR0
	movff	PROC_PTR+1,FSR0H
	movff	PROC_PTR+0,FSR0L
; process a message
	call	MSG_PROC_MSG
; exit procedure if no byte was processed
	addlw	0			; update ZERO flag
	btfsc	STATUS,Z
	return				; exit procedure if W is 0
; decrese PROC_LEN
	subwf	PROC_LEN
; advance PROC_PTR
	addwf	PROC_PTR+0
	movlw	0
	addwfc	PROC_PTR+1
; exit procedure if no bytes left
	movf	PROC_LEN
	btfsc	STATUS,Z
	return				; exit procedure if PROC_LEN is 0
; process next message
	bra	MSG_PROC

;-------------------------------------------------------------------------------------------
; process a message
; input:	FSR0		address of next byte to process
;		PROC_LEN	length of data to process
; output:	WREG		number of bytes processed
; changes:	PROC_LEN, FSR0, FSR1, PROD, X?_REG, WREG
;-------------------------------------------------------------------------------------------
MSG_PROC_MSG:
; calculate and jump to address in jump table
	movlw	4
	mulwf	INDF0
	movlw	low MSG_PROC_MSG_TAB
	addwf	PRODL
	movlw	high MSG_PROC_MSG_TAB
	addwfc	PRODH,W
	movwf	PCLATH
	clrf	PCLATU
	movf	PRODL,W
	movwf	PCL

;-------------------------------------------------------------------------------------------
; jump table for message processing
;-------------------------------------------------------------------------------------------
MSG_PROC_MSG_TAB:
	goto	MSG_SET_OUT_A		; 0x00
	goto	MSG_SET_OUT_B		; 0x01
	goto	MSG_SET_OUT_A_B		; 0x02
	goto	MSG_INVAL		; 0x03
	goto	MSG_READ_IN_A		; 0x04
	goto	MSG_READ_IN_B		; 0x05
	goto	MSG_READ_IN_A_B		; 0x06
	goto	MSG_INVAL		; 0x07
	goto	MSG_BURST_OUT_A		; 0x08
	goto	MSG_BURST_OUT_B		; 0x09
	goto	MSG_BURST_OUT_A_B	; 0x0A
	goto	MSG_INVAL		; 0x0B
	goto	MSG_BURST_OUT_IN_A	; 0x0C
	goto	MSG_BURST_OUT_IN_B	; 0x0D
	goto	MSG_BURST_OUT_IN_A_B	; 0x0E
	goto	MSG_INVAL		; 0x0F
	goto	MSG_SET_PWM_A		; 0x10
	goto	MSG_SET_PWM_B		; 0x11
	goto	MSG_SET_PWM_A_B		; 0x12
	goto	MSG_SET_PWM_SPEED	; 0x13
	goto	MSG_INVAL		; 0x14
	goto	MSG_INVAL		; 0x15
	goto	MSG_INVAL		; 0x16
	goto	MSG_INVAL		; 0x17
	goto	MSG_READ_ANA_A		; 0x18
	goto	MSG_READ_ANA_B		; 0x19
	goto	MSG_READ_ANA_C		; 0x1A
	goto	MSG_READ_ANA_D		; 0x1B
	goto	MSG_READ_ANA_E		; 0x1C
	goto	MSG_READ_ANA_A_E	; 0x1D
	goto	MSG_INVAL		; 0x1E
	goto	MSG_INVAL		; 0x1F
	goto	MSG_SET_BAUD		; 0x20
	goto	MSG_SER_SEND		; 0x21
	goto	MSG_INVAL		; 0x22
	goto	MSG_INVAL		; 0x23
	goto	MSG_INVAL		; 0x24
	goto	MSG_INVAL		; 0x25
	goto	MSG_INVAL		; 0x26
	goto	MSG_INVAL		; 0x27
	goto	MSG_INVAL		; 0x28
	goto	MSG_INVAL		; 0x29
	goto	MSG_INVAL		; 0x2A
	goto	MSG_INVAL		; 0x2B
	goto	MSG_INVAL		; 0x2C
	goto	MSG_INVAL		; 0x2D
	goto	MSG_INVAL		; 0x2E
	goto	MSG_INVAL		; 0x2F
	goto	MSG_I2C_SPEED		; 0x30
	goto	MSG_I2C_START		; 0x31
	goto	MSG_I2C_WRITE		; 0x32
	goto	MSG_I2C_READ		; 0x33
	goto	MSG_I2C_ACK		; 0x34
	goto	MSG_I2C_NAK		; 0x35
	goto	MSG_I2C_STOP		; 0x36
	goto	MSG_INVAL		; 0x37
	goto	MSG_INVAL		; 0x38
	goto	MSG_INVAL		; 0x39
	goto	MSG_INVAL		; 0x3A
	goto	MSG_INVAL		; 0x3B
	goto	MSG_INVAL		; 0x3C
	goto	MSG_INVAL		; 0x3D
	goto	MSG_INVAL		; 0x3E
	goto	MSG_INVAL		; 0x3F
	goto	MSG_INVAL		; 0x40
	goto	MSG_INVAL		; 0x41
	goto	MSG_INVAL		; 0x42
	goto	MSG_INVAL		; 0x43
	goto	MSG_INVAL		; 0x44
	goto	MSG_INVAL		; 0x45
	goto	MSG_INVAL		; 0x46
	goto	MSG_INVAL		; 0x47
	goto	MSG_INVAL		; 0x48
	goto	MSG_INVAL		; 0x49
	goto	MSG_INVAL		; 0x4A
	goto	MSG_INVAL		; 0x4B
	goto	MSG_INVAL		; 0x4C
	goto	MSG_INVAL		; 0x4D
	goto	MSG_INVAL		; 0x4E
	goto	MSG_INVAL		; 0x4F
	goto	MSG_INVAL		; 0x50
	goto	MSG_INVAL		; 0x51
	goto	MSG_INVAL		; 0x52
	goto	MSG_INVAL		; 0x53
	goto	MSG_INVAL		; 0x54
	goto	MSG_INVAL		; 0x55
	goto	MSG_INVAL		; 0x56
	goto	MSG_INVAL		; 0x57
	goto	MSG_INVAL		; 0x58
	goto	MSG_INVAL		; 0x59
	goto	MSG_INVAL		; 0x5A
	goto	MSG_INVAL		; 0x5B
	goto	MSG_INVAL		; 0x5C
	goto	MSG_INVAL		; 0x5D
	goto	MSG_INVAL		; 0x5E
	goto	MSG_INVAL		; 0x5F
	goto	MSG_INVAL		; 0x60
	goto	MSG_INVAL		; 0x61
	goto	MSG_INVAL		; 0x62
	goto	MSG_INVAL		; 0x63
	goto	MSG_INVAL		; 0x64
	goto	MSG_INVAL		; 0x65
	goto	MSG_INVAL		; 0x66
	goto	MSG_INVAL		; 0x67
	goto	MSG_INVAL		; 0x68
	goto	MSG_INVAL		; 0x69
	goto	MSG_INVAL		; 0x6A
	goto	MSG_INVAL		; 0x6B
	goto	MSG_INVAL		; 0x6C
	goto	MSG_INVAL		; 0x6D
	goto	MSG_INVAL		; 0x6E
	goto	MSG_INVAL		; 0x6F
	goto	MSG_INVAL		; 0x70
	goto	MSG_INVAL		; 0x71
	goto	MSG_INVAL		; 0x72
	goto	MSG_INVAL		; 0x73
	goto	MSG_INVAL		; 0x74
	goto	MSG_INVAL		; 0x75
	goto	MSG_INVAL		; 0x76
	goto	MSG_INVAL		; 0x77
	goto	MSG_INVAL		; 0x78
	goto	MSG_INVAL		; 0x79
	goto	MSG_INVAL		; 0x7A
	goto	MSG_INVAL		; 0x7B
	goto	MSG_INVAL		; 0x7C
	goto	MSG_INVAL		; 0x7D
	goto	MSG_INVAL		; 0x7E
	goto	MSG_INVAL		; 0x7F
	goto	MSG_INVAL		; 0x80
	goto	MSG_INVAL		; 0x81
	goto	MSG_INVAL		; 0x82
	goto	MSG_INVAL		; 0x83
	goto	MSG_INVAL		; 0x84
	goto	MSG_INVAL		; 0x85
	goto	MSG_INVAL		; 0x86
	goto	MSG_INVAL		; 0x87
	goto	MSG_INVAL		; 0x88
	goto	MSG_INVAL		; 0x89
	goto	MSG_INVAL		; 0x8A
	goto	MSG_INVAL		; 0x8B
	goto	MSG_INVAL		; 0x8C
	goto	MSG_INVAL		; 0x8D
	goto	MSG_INVAL		; 0x8E
	goto	MSG_INVAL		; 0x8F
	goto	MSG_INVAL		; 0x90
	goto	MSG_INVAL		; 0x91
	goto	MSG_INVAL		; 0x92
	goto	MSG_INVAL		; 0x93
	goto	MSG_INVAL		; 0x94
	goto	MSG_INVAL		; 0x95
	goto	MSG_INVAL		; 0x96
	goto	MSG_INVAL		; 0x97
	goto	MSG_INVAL		; 0x98
	goto	MSG_INVAL		; 0x99
	goto	MSG_INVAL		; 0x9A
	goto	MSG_INVAL		; 0x9B
	goto	MSG_INVAL		; 0x9C
	goto	MSG_INVAL		; 0x9D
	goto	MSG_INVAL		; 0x9E
	goto	MSG_INVAL		; 0x9F
	goto	MSG_INVAL		; 0xA0
	goto	MSG_INVAL		; 0xA1
	goto	MSG_INVAL		; 0xA2
	goto	MSG_INVAL		; 0xA3
	goto	MSG_INVAL		; 0xA4
	goto	MSG_INVAL		; 0xA5
	goto	MSG_INVAL		; 0xA6
	goto	MSG_INVAL		; 0xA7
	goto	MSG_INVAL		; 0xA8
	goto	MSG_INVAL		; 0xA9
	goto	MSG_INVAL		; 0xAA
	goto	MSG_INVAL		; 0xAB
	goto	MSG_INVAL		; 0xAC
	goto	MSG_INVAL		; 0xAD
	goto	MSG_INVAL		; 0xAE
	goto	MSG_INVAL		; 0xAF
	goto	MSG_INVAL		; 0xB0
	goto	MSG_INVAL		; 0xB1
	goto	MSG_INVAL		; 0xB2
	goto	MSG_INVAL		; 0xB3
	goto	MSG_INVAL		; 0xB4
	goto	MSG_INVAL		; 0xB5
	goto	MSG_INVAL		; 0xB6
	goto	MSG_INVAL		; 0xB7
	goto	MSG_INVAL		; 0xB8
	goto	MSG_INVAL		; 0xB9
	goto	MSG_INVAL		; 0xBA
	goto	MSG_INVAL		; 0xBB
	goto	MSG_INVAL		; 0xBC
	goto	MSG_INVAL		; 0xBD
	goto	MSG_INVAL		; 0xBE
	goto	MSG_INVAL		; 0xBF
	goto	MSG_INVAL		; 0xC0
	goto	MSG_INVAL		; 0xC1
	goto	MSG_INVAL		; 0xC2
	goto	MSG_INVAL		; 0xC3
	goto	MSG_INVAL		; 0xC4
	goto	MSG_INVAL		; 0xC5
	goto	MSG_INVAL		; 0xC6
	goto	MSG_INVAL		; 0xC7
	goto	MSG_INVAL		; 0xC8
	goto	MSG_INVAL		; 0xC9
	goto	MSG_INVAL		; 0xCA
	goto	MSG_INVAL		; 0xCB
	goto	MSG_INVAL		; 0xCC
	goto	MSG_INVAL		; 0xCD
	goto	MSG_INVAL		; 0xCE
	goto	MSG_INVAL		; 0xCF
	goto	MSG_INVAL		; 0xD0
	goto	MSG_INVAL		; 0xD1
	goto	MSG_INVAL		; 0xD2
	goto	MSG_INVAL		; 0xD3
	goto	MSG_INVAL		; 0xD4
	goto	MSG_INVAL		; 0xD5
	goto	MSG_INVAL		; 0xD6
	goto	MSG_INVAL		; 0xD7
	goto	MSG_INVAL		; 0xD8
	goto	MSG_INVAL		; 0xD9
	goto	MSG_INVAL		; 0xDA
	goto	MSG_INVAL		; 0xDB
	goto	MSG_INVAL		; 0xDC
	goto	MSG_INVAL		; 0xDD
	goto	MSG_INVAL		; 0xDE
	goto	MSG_INVAL		; 0xDF
	goto	MSG_INVAL		; 0xE0
	goto	MSG_INVAL		; 0xE1
	goto	MSG_INVAL		; 0xE2
	goto	MSG_INVAL		; 0xE3
	goto	MSG_INVAL		; 0xE4
	goto	MSG_INVAL		; 0xE5
	goto	MSG_INVAL		; 0xE6
	goto	MSG_INVAL		; 0xE7
	goto	MSG_INVAL		; 0xE8
	goto	MSG_INVAL		; 0xE9
	goto	MSG_INVAL		; 0xEA
	goto	MSG_INVAL		; 0xEB
	goto	MSG_INVAL		; 0xEC
	goto	MSG_INVAL		; 0xED
	goto	MSG_INVAL		; 0xEE
	goto	MSG_INVAL		; 0xEF
	goto	MSG_GET_ID		; 0xF0
	goto	MSG_SET_ID		; 0xF1
	goto	MSG_INVAL		; 0xF2
	goto	MSG_INVAL		; 0xF3
	goto	MSG_INVAL		; 0xF4
	goto	MSG_INVAL		; 0xF5
	goto	MSG_INVAL		; 0xF6
	goto	MSG_INVAL		; 0xF7
	goto	MSG_INVAL		; 0xF8
	goto	MSG_INVAL		; 0xF9
	goto	MSG_INVAL		; 0xFA
	goto	MSG_INVAL		; 0xFB
	goto	MSG_INVAL		; 0xFC
	goto	MSG_INVAL		; 0xFD
	goto	MSG_INVAL		; 0xFE
	goto	MSG_READ_VER		; 0xFF

;-------------------------------------------------------------------------------------------
; 0x00 <data>
; set output port A
;-------------------------------------------------------------------------------------------
MSG_SET_OUT_A:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new value to output port A
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	setf	PORT_BUS_T		; set data bus port back to input
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x01 <data>
; set output port B
;-------------------------------------------------------------------------------------------
MSG_SET_OUT_B:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new value to output port B
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x02 <data A> <data B>
; set output ports A and B
;-------------------------------------------------------------------------------------------
MSG_SET_OUT_A_B:
; check length
	movlw	3-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new value to output ports A and B
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	movff	PREINC0,PORT_BUS_O	; output value for output port B on data bus
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	3

;-------------------------------------------------------------------------------------------
; 0x04
; read input port A
;-------------------------------------------------------------------------------------------
MSG_READ_IN_A:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x04			; put data into send buffer
	movwf	POSTINC0		; (value of port will be written below)
; read value from input port A
	bcf	PIN_IN_A_nOE		; set input port A output enable
	movff	PORT_BUS_I,POSTINC0	; read value and put it into send buffer
	bsf	PIN_IN_A_nOE		; clear input port A output enable
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x05
; read input port B
;-------------------------------------------------------------------------------------------
MSG_READ_IN_B:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x05			; put data into send buffer
	movwf	POSTINC0		; (value of port will be written below)
; read value from input port B
	bcf	PIN_IN_B_nOE		; set input port B output enable
	movff	PORT_BUS_I,POSTINC0	; read value and put it into send buffer
	bsf	PIN_IN_B_nOE		; clear input port B output enable
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x06
; read input ports A and B
;-------------------------------------------------------------------------------------------
MSG_READ_IN_A_B:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x06			; put data into send buffer
	movwf	POSTINC0		; (values of ports will be written below)
; read value from input port A
	bcf	PIN_IN_A_nOE		; set input port A output enable
	movff	PORT_BUS_I,POSTINC0	; read value and put it into send buffer
	bsf	PIN_IN_A_nOE		; clear input port A output enable
; read value from input port B
	bcf	PIN_IN_B_nOE		; set input port B output enable
	movff	PORT_BUS_I,POSTINC0	; read value and put it into send buffer
	bsf	PIN_IN_B_nOE		; clear input port B output enable
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x08 <data 0> ... <data 15>
; burst output on output port A
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_A:
; check length
	movlw	17-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write 16 new values to output port A
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_A_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	decfsz	A1_REG
	bra	MSG_BURST_OUT_A_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	17

;-------------------------------------------------------------------------------------------
; 0x09 <data 0> ... <data 15>
; burst output on output port B
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_B:
; check length
	movlw	17-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write 16 new values to output port A
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_B_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port B on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	decfsz	A1_REG
	bra	MSG_BURST_OUT_B_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	17

;-------------------------------------------------------------------------------------------
; 0x0A <data A 0> <data B 0> ... <data A 15> <data B 15>
; burst output on output ports A and B
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_A_B:
; check length
	movlw	33-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write 16 new values to output port A and output port B
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_A_B_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	movff	PREINC0,PORT_BUS_O	; output value for output port B on data bus
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	decfsz	A1_REG
	bra	MSG_BURST_OUT_A_B_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	33

;-------------------------------------------------------------------------------------------
; 0x0C <data 0> ... <data 15>
; burst output on output port A combined with input on input port A
; (after each byte is output, a byte is input)
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_IN_A:
; check length
	movlw	17-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-17+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR1,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
	movlw	17			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x0C			; put data into send buffer
	movwf	POSTINC1		; (value of port will be written below)
; write to output port A and read from input port A 16 times
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_IN_A_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	bcf	PIN_IN_A_nOE		; set input port A output enable
	movff	PORT_BUS_I,POSTINC1	; read value and put it into send buffer
	bsf	PIN_IN_A_nOE		; clear input port A output enable
	decfsz	A1_REG
	bra	MSG_BURST_OUT_IN_A_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	17

;-------------------------------------------------------------------------------------------
; 0x0D <data 0> ... <data 15>
; burst output on output port B combined with input on input port B
; (after each byte is output, a byte is input)
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_IN_B:
; check length
	movlw	17-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-17+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR1,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
	movlw	17			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x0D			; put data into send buffer
	movwf	POSTINC1		; (value of port will be written below)
; write to output port B and read from input port B 16 times
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_IN_B_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port B on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	bcf	PIN_IN_B_nOE		; set input port B output enable
	movff	PORT_BUS_I,POSTINC1	; read value and put it into send buffer
	bsf	PIN_IN_B_nOE		; clear input port B output enable
	decfsz	A1_REG
	bra	MSG_BURST_OUT_IN_B_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	17

;-------------------------------------------------------------------------------------------
; 0x0E <data A 0> <data B 0> ... <data A 15> <data B 15>
; burst output on output ports A and B combined with input on input port A and B
; (after each byte is output, a byte is input)
;-------------------------------------------------------------------------------------------
MSG_BURST_OUT_IN_A_B:
; check length
	movlw	33-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-33+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR1,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR1L
	movlw	0
	addwfc	FSR1H
	movlw	33			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x0E			; put data into send buffer
	movwf	POSTINC1		; (value of port will be written below)
; write to output ports A and B and read from input ports A and B 16 times
	movlw	16
	movwf	A1_REG
MSG_BURST_OUT_IN_A_B_LOOP:
	movff	PREINC0,PORT_BUS_O	; output value for output port A on data bus
	clrf	PORT_BUS_T		; set data bus port to output
	bsf	PIN_OUT_A_LE		; set output port A latch enable
	bcf	PIN_OUT_A_LE		; clear output port A latch enable
	movff	PREINC0,PORT_BUS_O	; output value for output port B on data bus
	bsf	PIN_OUT_B_LE		; set output port B latch enable
	bcf	PIN_OUT_B_LE		; clear output port B latch enable
	setf	PORT_BUS_T		; set data bus port back to input
	bcf	PIN_IN_A_nOE		; set input port A output enable
	movff	PORT_BUS_I,POSTINC1	; read value and put it into send buffer
	bsf	PIN_IN_A_nOE		; clear input port A output enable
	bcf	PIN_IN_B_nOE		; set input port B output enable
	movff	PORT_BUS_I,POSTINC1	; read value and put it into send buffer
	bsf	PIN_IN_B_nOE		; clear input port B output enable
	decfsz	A1_REG
	bra	MSG_BURST_OUT_IN_A_B_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	33

;-------------------------------------------------------------------------------------------
; 0x10 <data_h> <data_l>
; set PWM output A
;-------------------------------------------------------------------------------------------
MSG_SET_PWM_A:
; check length
	movlw	3-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new value to PWM output A
	movff	PREINC0,CCPR1L		; write value into PWM1 registers
	bcf	CCP1CON,DC1B1
	btfsc	PREINC0,7
	bsf	CCP1CON,DC1B1
	bcf	CCP1CON,DC1B0
	btfsc	INDF0,6
	bsf	CCP1CON,DC1B0
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	3

;-------------------------------------------------------------------------------------------
; 0x11 <data_h> <data_l>
; set PWM output B
;-------------------------------------------------------------------------------------------
MSG_SET_PWM_B:
; check length
	movlw	3-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new value to PWM output B
	movff	PREINC0,CCPR2L		; write value into PWM2 registers
	bcf	CCP2CON,DC2B1
	btfsc	PREINC0,7
	bsf	CCP2CON,DC2B1
	bcf	CCP2CON,DC2B0
	btfsc	INDF0,6
	bsf	CCP2CON,DC2B0
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	3

;-------------------------------------------------------------------------------------------
; 0x12 <data_h A> <data_l A> <data_h B> <data_l B>
; set PWM outputs A and B
;-------------------------------------------------------------------------------------------
MSG_SET_PWM_A_B:
; check length
	movlw	5-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; write new values to PWM outputs A and B
	movff	PREINC0,CCPR1L		; write value into PWM1 registers
	bcf	CCP1CON,DC1B1
	btfsc	PREINC0,7
	bsf	CCP1CON,DC1B1
	bcf	CCP1CON,DC1B0
	btfsc	INDF0,6
	bsf	CCP1CON,DC1B0
	movff	PREINC0,CCPR2L		; write value into PWM2 registers
	bcf	CCP2CON,DC2B1
	btfsc	PREINC0,7
	bsf	CCP2CON,DC2B1
	bcf	CCP2CON,DC2B0
	btfsc	INDF0,6
	bsf	CCP2CON,DC2B0
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	5

;-------------------------------------------------------------------------------------------
; 0x13 <speed>
; set PWM speed (i.e. frequency)
; speed: frequency in 250 Hz steps
;-------------------------------------------------------------------------------------------
MSG_SET_PWM_SPEED:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; set new PWM speed
	movlw	20			; 0..19 (0..4.75kHz) => slow (2.5 kHz)
	cpfslt	PREINC0
	bra	MSG_SET_PWM_SPEED_MEDIUM
	movlw	b'00000110'		; timer 2 on, prescaler 1:16, postscaler 1:1
	movwf	T2CON			; (PR2 = 249: 16*250=4000 cycles, period = 2.5 kHz at 40 MHz)
	bra	MSG_SET_PWM_SPEED_END
MSG_SET_PWM_SPEED_MEDIUM:
	movlw	80			; 20..79 (5..19.75kHz) => medium (10 kHz)
	cpfslt	INDF0
	bra	MSG_SET_PWM_SPEED_FAST
	movlw	b'00000101'		; timer 2 on, prescaler 1:4, postscaler 1:1
	movwf	T2CON			; (PR2 = 249: 4*250=1000 cycles, period = 10 kHz at 40 MHz)
	bra	MSG_SET_PWM_SPEED_END
MSG_SET_PWM_SPEED_FAST:
					; 80..255 (20..63.75kHz) => fast (40 kHz)
	movlw	b'00000100'		; timer 2 on, prescaler 1:1, postscaler 1:1
	movwf	T2CON			; (PR2 = 249: 1*250=250 cycles, period = 40 kHz at 40 MHz)
MSG_SET_PWM_SPEED_END:
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x18
; read analog input A
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_A:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x18			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog input A
	movlw	b'11000111'		; select analog input AN0
	andwf	ADCON0
	movlw	b'00000000'
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_A_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_A_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_A_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_A_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_A_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_A_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x19
; read analog input B
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_B:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x19			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog input B
	movlw	b'11000111'		; select analog input AN1
	andwf	ADCON0
	movlw	b'00001000'
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_B_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_B_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_B_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_B_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_B_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_B_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x1A
; read analog input C
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_C:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x1A			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog input C
	movlw	b'11000111'		; select analog input AN2
	andwf	ADCON0
	movlw	b'00010000'
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_C_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_C_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_C_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_C_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_C_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_C_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x1B
; read analog input D
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_D:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x1B			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog input D
	movlw	b'11000111'		; select analog input AN3
	andwf	ADCON0
	movlw	b'00011000'
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_D_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_D_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_D_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_D_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_D_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_D_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x1C
; read analog input E
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_E:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-3+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	3			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x1C			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog input E
	movlw	b'11000111'		; select analog input AN4
	andwf	ADCON0
	movlw	b'00100000'
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_E_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_E_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_E_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_E_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_E_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_E_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x1D
; read analog inputs A ... E
;-------------------------------------------------------------------------------------------
MSG_READ_ANA_A_E:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-11+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	11			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x1D			; put data into send buffer
	movwf	POSTINC0		; (value of analog input will be written below)
; read value from analog inputs
	movlw	5			; read 5 analog inputs
	movwf	B1_REG
	movlw	b'00000000'		; start with AN0
	movwf	C1_REG
MSG_READ_ANA_A_E_LOOP:
	movlw	b'11000111'		; select analog input
	andwf	ADCON0
	movf	C1_REG,W
	iorwf	ADCON0
	movlw	50			; time for input capacitor to load (15us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_A_E_LOAD:
	decfsz	A1_REG
	bra	MSG_READ_ANA_A_E_LOAD
	bsf	ADCON0,GO		; start conversion (12 steps, 1.6us each => 19.2us)
	movlw	50			; wait for conversion to finish (25us at 40MHz)
	movwf	A1_REG
MSG_READ_ANA_A_E_WAIT:
	btfss	ADCON0,GO
	bra	MSG_READ_ANA_A_E_DONE	; end loop if conversion is done
	decfsz	A1_REG			; wait loop
	bra	MSG_READ_ANA_A_E_WAIT
	bsf	ADRESL,0		; set last bit to indicate timeout (last 6 bits are always 0)
MSG_READ_ANA_A_E_DONE:
	movff	ADRESH,POSTINC0		; write high part of result into send buffer
	movff	ADRESL,POSTINC0		; write low part of result into send buffer
	movlw	b'00001000'		; increment analog input number
	addwf	C1_REG
	decfsz	B1_REG			; loop
	bra	MSG_READ_ANA_A_E_LOOP
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x20 <bit_time_h> <bit_time_l>
; set baudrate
; bit_time: time for a single bit (in 100 ns)
;-------------------------------------------------------------------------------------------
MSG_SET_BAUD:
; check length
	movlw	3-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; get bit timing into A2_REG
	movff	PREINC0,A2_REG+1
	movff	PREINC0,A2_REG+0
; set baudrate of serial port
	call	SER_SET_BAUD
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; return number of bytes processed
	retlw	3

;-------------------------------------------------------------------------------------------
; 0x21 <byte>
; send a byte using the serial port
;-------------------------------------------------------------------------------------------
MSG_SER_SEND:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; get data byte into A1_REG
	movff	PREINC0,A1_REG
; send byte
	call	SER_SEND
	btfsc	STATUS,C		; if buffer was full, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x30 <bit_time>
; set speed of I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_SPEED:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; get bit time into A1_REG
	movff	PREINC0,A1_REG
; set speed
	call	I2C_SPEED
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x31
; output start condition on I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_START:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; output start condition
	call	I2C_STA
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x32 <data>
; write a data byte to I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_WRITE:
; check length
	movlw	2-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; get data byte into A1_REG
	movff	PREINC0,A1_REG
; perhaps jump to second part
	btfsc	MSG_BITS_I2C_PART2
	bra	MSG_I2C_WRITE_PART2
; start writing
	call	I2C_WRITE_START
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; remember that first step of write is already processed
	bsf	MSG_BITS_I2C_PART2
; process second part later
	retlw	0
; second part
MSG_I2C_WRITE_PART2:
; check if enough space in send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; get acknowledge state of write
	call	I2C_WRITE_RESULT
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x32			; put data into send buffer
	movwf	POSTINC0
	movf	A1_REG,W
	movwf	POSTINC0
; second part of write is processed
	bcf	MSG_BITS_I2C_PART2
; return number of bytes processed
	retlw	2

;-------------------------------------------------------------------------------------------
; 0x33
; read a data byte from I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_READ:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; perhaps jump to second part
	btfsc	MSG_BITS_I2C_PART2
	bra	MSG_I2C_READ_PART2
; start reading
	call	I2C_READ_START
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; remember that first step of read is already processed
	bsf	MSG_BITS_I2C_PART2
; process second part later
	retlw	0
; second part
MSG_I2C_READ_PART2:
; check if enough space in send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; get byte read from I2C
	call	I2C_READ_RESULT
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0x33			; put data into send buffer
	movwf	POSTINC0
	movf	A1_REG,W
	movwf	POSTINC0
; second part of read is processed
	bcf	MSG_BITS_I2C_PART2
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x34
; output acknowledge on I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_ACK:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; output acknowledge
	call	I2C_ACK
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x35
; output not-acknowledge on I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_NAK:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; output not-acknowledge
	call	I2C_NAK
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x36
; output stop condition on I2C bus
;-------------------------------------------------------------------------------------------
MSG_I2C_STOP:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; output stop condition
	call	I2C_STO
	btfsc	STATUS,C		; if bus was busy, ...
	retlw	0			; ..., process message again later
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0xF0
; get ID of this hacker port
;-------------------------------------------------------------------------------------------
MSG_GET_ID:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-17+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	17			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0xF0			; put data into send buffer
	movwf	POSTINC0
	call	EE_GET_ID		; read ID from EEPROM
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0xF1 <ID0> <ID1> ... <ID15>
; set ID of this hacker port
;-------------------------------------------------------------------------------------------
MSG_SET_ID:
; check length
	movlw	17-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-2+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; if setting identification is disabled
	btfss	MSG_SET_ID_LOCK
	bra	MSG_SET_ID_OK
	setf	C1_REG			; set error code to "setting identification disabled"
	bra	MSG_SET_ID_END
; write ID to EEPROM
MSG_SET_ID_OK:
	movf	POSTINC0,W		; move to new ID
	call	EE_SET_ID		; write ID to EEPROM
	clrf	C1_REG			; set error code to "success"
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; put answer message into send buffer
MSG_SET_ID_END:
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	2			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0xF1			; put data into send buffer
	movwf	POSTINC0
	movff	C1_REG,POSTINC0		; error code
; return number of bytes processed
	retlw	17

;-------------------------------------------------------------------------------------------
; 0xFF
; read version information
;-------------------------------------------------------------------------------------------
MSG_READ_VER:
; check length
	movlw	1-1
	cpfsgt	PROC_LEN
	bra	MSG_INVAL		; jump if length is too small
; check if enough space in send buffer
	movlw	63-9+1
	cpfslt	USB_Sx_CNT		; if send buffer too full, ...
	retlw	0			; ..., process this message again later
; put answer message into send buffer
	lfsr	FSR0,USB_Sx_BUF		; calculate address in send buffer
	movf	USB_Sx_CNT,W
	addwf	FSR0L
	movlw	0
	addwfc	FSR0H
	movlw	9			; increase length of data in send buffer
	addwf	USB_Sx_CNT
	movlw	0xFF			; put data into send buffer
	movwf	POSTINC0
	movlw	0x00			; variant 0x00: original hacker port with PIC18F442 and USBN9604
	movwf	POSTINC0
	movlw	VER_MAJ			; major version
	movwf	POSTINC0
	movlw	VER_MIN			; minor version
	movwf	POSTINC0
	movlw	VER_REV			; revision
	movwf	POSTINC0
	movlw	VER_YEAR_H		; year_h of release date
	movwf	POSTINC0
	movlw	VER_YEAR_L		; year_l of release date
	movwf	POSTINC0
	movlw	VER_MONTH		; month of release date
	movwf	POSTINC0
	movlw	VER_DAY			; day of release date
	movwf	POSTINC0
; return number of bytes processed
	retlw	1

;-------------------------------------------------------------------------------------------
; 0x??
; invalid message
;-------------------------------------------------------------------------------------------
MSG_INVAL:
; disable setting of identification
	bsf	MSG_SET_ID_LOCK
; discard rest of packet - return number of bytes left in packet
	movf	PROC_LEN,W
	return



;===========================================================================================
; global variables
;===========================================================================================



;===========================================================================================
; global procedures
;===========================================================================================

	global		MSG_INIT, MSG_PROC



	end
